#!/usr/bin/env python
#
# SPDX-License-Identifier: GPL-3.0+

"""
Simple self tests for fwhunt_scan
"""

import unittest

from .uefi_protocols import PROTOCOLS_GUIDS, GUID_FROM_VALUE


class TestInternal(unittest.TestCase):
    """internal tests of privaet API"""

    def test_guid_convert(self):
        """convert to GUID by index and value"""

        self.assertEqual(
            PROTOCOLS_GUIDS[144].value,
            "5C6FA2C9-9768-45F6-8E645AECCADAB481",
        )
        self.assertEqual(
            PROTOCOLS_GUIDS[144].bytes,
            b"\xc9\xa2o\\h\x97\xf6E\x8edZ\xec\xca\xda\xb4\x81",
        )
        self.assertEqual(
            GUID_FROM_VALUE["C2702B74-800C-4131-87468FB5B89CE4AC"].name,
            "EFI_SMM_ACCESS2_PROTOCOL_GUID",
        )


if __name__ == "__main__":
    unittest.main()



import glob
from typing import List

import click

from fwhunt_scan import UefiAnalyzer, UefiRule, UefiScanner

# images = glob.glob("test/S15_02050100.bin.out/*")
images = glob.glob("test/S74_01070000.bin.out/*")
rules_yml = glob.glob("rules/*.yml")

prefix = click.style("Scanner result", fg="green")
no_threat = click.style("No threat detected", fg="green")
threat = click.style("FwHunt rule has been triggered and threat detected!", fg="red")

# init rules
uefi_rules_all: List[UefiRule] = list()
for rule in rules_yml:
    with open(rule, "r") as f:
        rule_content = f.read()

    rule = UefiRule(rule_content=rule_content)
    uefi_rules_all.append(rule)

for image_path in images:

    uefi_rules: List[UefiRule] = list()
    for rule in uefi_rules_all:
        for guid in rule.volume_guids:
            if guid.lower() in image_path:
                uefi_rules.append(rule)

    if not uefi_rules:
        continue

    # init analyzer
    uefi_analyzer = UefiAnalyzer(image_path=image_path)

    # init scanner
    scanner = UefiScanner(uefi_analyzer, uefi_rules)
    for result in scanner.results:
        print(result.rule.name, result.variant_label, result.res)

    uefi_analyzer.close()



#!/usr/bin/env python3
#
# SPDX-License-Identifier: GPL-3.0+

import os
from typing import List

import click

TAG = "fwhunt_scan"


@click.group()
def cli():
    pass


@click.command()
def build():
    """Build docker image."""

    os.system(" ".join(["docker", "build", "-t", TAG, "."]))


@click.command()
@click.argument("module_path")
def analyze_module(module_path: str) -> bool:
    """Analyze single UEFI module."""

    if not os.path.isfile(module_path):
        print("{} check module path".format(click.style("ERROR", fg="red", bold=True)))
        return False

    fpath = os.path.realpath(module_path)

    cmdstr = " ".join(
        [
            "docker",
            "run",
            "--rm",
            "-it",
            "-v",
            f"{fpath}:/tmp/module:ro",
            TAG,
            "analyze-module",
            "/tmp/module",
        ]
    )

    print(f"Command: {cmdstr}")

    os.system(cmdstr)

    return True


@click.command()
@click.argument("module_path")
@click.option("-r", "--rule", help="The path to the rule.", multiple=True)
def scan_module(module_path: str, rule: List[str]) -> bool:
    """Scan singe UEFI module."""

    rules = rule

    if not os.path.isfile(module_path):
        print("{} check module path".format(click.style("ERROR", fg="red", bold=True)))
        return False
    if not all(rule and os.path.isfile(rule) for rule in rules):
        print("{} check rule(s) path".format(click.style("ERROR", fg="red", bold=True)))
        return False

    cmd = [
        "docker",
        "run",
        "--rm",
        "-it",
        "-v",
        f"{os.path.realpath(module_path)}:/tmp/module:ro",
    ]
    rules_cmd = [TAG, "scan-module", "/tmp/module"]
    for r in rules:
        _, name = os.path.split(r)
        cmd += ["-v", f"{os.path.realpath(r)}:/tmp/{name}:ro"]
        rules_cmd += ["-r", f"/tmp/{name}"]

    cmd += rules_cmd
    cmdstr = " ".join(cmd)

    print(f"Commamd: {cmdstr}")

    os.system(cmdstr)

    return True


@click.command()
@click.argument("image_path")
@click.option("-r", "--rule", help="The path to the rule.", multiple=True)
@click.option("-d", "--rules_dir", help="The path to the rules directory.")
def scan_firmware(image_path: str, rule: List[str], rules_dir: str) -> bool:
    """Scan UEFI firmware image."""

    rules = rule

    if not os.path.isfile(image_path):
        print("{} check image path".format(click.style("ERROR", fg="red", bold=True)))
        return False
    if not (rules_dir or rule):
        print("{} check command".format(click.style("ERROR", fg="red", bold=True)))
        return False
    if rules_dir and not os.path.isdir(rules_dir):
        print(
            "{} check rules directory path".format(
                click.style("ERROR", fg="red", bold=True)
            )
        )
        return False
    if rule and not all(rule and os.path.isfile(rule) for rule in rules):
        print("{} check rule(s) path".format(click.style("ERROR", fg="red", bold=True)))
        return False

    cmd = [
        "docker",
        "run",
        "--rm",
        "-it",
        "-v",
        f"{os.path.realpath(image_path)}:/tmp/image:ro",
    ]
    rules_cmd = [TAG, "scan-firmware", "/tmp/image"]
    for r in rules:
        _, name = os.path.split(r)
        cmd += ["-v", f"{os.path.realpath(r)}:/tmp/{name}:ro"]
        rules_cmd += ["-r", f"/tmp/{name}"]

    if rules_dir:
        _, name = os.path.split(rules_dir)
        cmd += ["-v", f"{os.path.realpath(rules_dir)}:/tmp/{name}:ro"]
        rules_cmd += ["--rules_dir", f"/tmp/{name}"]

    cmd += rules_cmd
    cmdstr = " ".join(cmd)

    print(f"Commamd: {cmdstr}")

    os.system(cmdstr)

    return True


cli.add_command(build)
cli.add_command(analyze_module)
cli.add_command(scan_module)
cli.add_command(scan_firmware)

if __name__ == "__main__":
    cli()



#!/usr/bin/env python3
#
# SPDX-License-Identifier: GPL-3.0+
#
# pylint: disable=invalid-name,missing-module-docstring,missing-function-docstring

# fwhunt_scan: tools for analyzing UEFI firmware and checking UEFI modules with FwHunt rules

import json
import os
import pathlib
import tempfile
from typing import Dict, List

import click

from fwhunt_scan import UefiAnalyzer, UefiExtractor, UefiRule, UefiScanner


@click.group()
def cli():
    pass


@click.command()
@click.argument("module_path")
@click.option("-o", "--out", help="Output JSON file.")
def analyze_module(module_path: str, out: str) -> bool:
    """Analyze single UEFI module."""

    if not os.path.isfile(module_path):
        print("{} check module path".format(click.style("ERROR", fg="red", bold=True)))
        return False

    # on linux platforms you can pass blob via shm://
    # uefi_analyzer = UefiAnalyzer(blob=data)

    summary = None
    with UefiAnalyzer(image_path=module_path) as uefi_analyzer:
        summary = uefi_analyzer.get_summary()

    if out:
        with open(out, "w") as f:
            json.dump(summary, f, indent=4)
    else:
        print(json.dumps(summary, indent=4))

    return True


@click.command()
@click.argument("module_path")
@click.option("-r", "--rule", help="The path to the rule.", multiple=True)
def scan_module(module_path: str, rule: List[str]) -> bool:
    """Scan single UEFI module."""

    rules = rule

    if not os.path.isfile(module_path):
        print("{} check module path".format(click.style("ERROR", fg="red", bold=True)))
        return False
    if not all(rule and os.path.isfile(rule) for rule in rules):
        print("{} check rule(s) path".format(click.style("ERROR", fg="red", bold=True)))
        return False

    # on linux platforms you can pass blob via shm://
    # uefi_analyzer = UefiAnalyzer(blob=data)

    uefi_analyzer = UefiAnalyzer(image_path=module_path)

    uefi_rules: List[UefiRule] = list()

    for r in rules:
        with open(r, "r") as f:
            rule_content = f.read()
            uefi_rule = UefiRule(rule_content=rule_content)
            uefi_rules.append(uefi_rule)

    scanner = UefiScanner(uefi_analyzer, uefi_rules)
    prefix = click.style("Scanner result", fg="green")

    no_threat = click.style("No threat detected", fg="green")
    threat = click.style(
        "FwHunt rule has been triggered and threat detected!", fg="red"
    )

    for result in scanner.results:
        msg = threat if result.res else no_threat
        print(
            f"{prefix} {result.rule.name} (variant: {result.variant_label}) {msg} ({module_path})"
        )

    uefi_analyzer.close()

    return True


@click.command()
@click.argument("image_path")
@click.option("-r", "--rule", help="The path to the rule.", multiple=True)
@click.option("-d", "--rules_dir", help="The path to the rules directory.")
def scan_firmware(image_path: str, rule: List[str], rules_dir: str) -> bool:
    """Scan UEFI firmware image."""

    rules = list(rule)
    error_prefix = click.style("ERROR", fg="red", bold=True)
    if not rules_dir:
        if not all(rules and os.path.isfile(rule) for rule in rules):
            print(f"{error_prefix} check rule(s) path")
            return False
    else:
        rules += list(map(str, pathlib.Path(rules_dir).rglob("*.yml")))

    if not os.path.isfile(image_path):
        print(f"{error_prefix} check image path")
        return False

    prefix = click.style("Scanner result", fg="green")
    no_threat = click.style("No threat detected", fg="green")
    threat = click.style(
        "FwHunt rule has been triggered and threat detected!", fg="red"
    )

    # on linux platforms you can pass blob via shm://
    # uefi_analyzer = UefiAnalyzer(blob=data)

    uefi_rules: List[UefiRule] = list()

    for r in rules:
        with open(r, "r") as f:
            uefi_rules.append(UefiRule(rule_content=f.read()))

    # Select the rules with `target: firmware`
    uefi_rules_fw: List[UefiRule] = list(
        filter(lambda rule: rule.target == "firmware", uefi_rules)
    )
    if len(uefi_rules_fw):
        with UefiAnalyzer(image_path=image_path) as uefi_analyzer_fw:
            scanner_fw = UefiScanner(uefi_analyzer_fw, uefi_rules_fw)
            for result in scanner_fw.results:
                msg = threat if result.res else no_threat
                print(
                    f"{prefix} {result.rule.name} (variant: {result.variant_label}) {msg}"
                )

    # Group rules by guids
    rules_guids: Dict[str, List[UefiRule]] = dict()
    for uefi_rule in uefi_rules:
        if uefi_rule.volume_guids is None:
            print(f"[I] Specify volume_guids in {uefi_rule.name} or use scan command")
            continue
        for guid in [g.lower() for g in uefi_rule.volume_guids]:
            lower_guid = guid.lower()
            if lower_guid not in rules_guids:
                rules_guids[lower_guid] = list()
            rules_guids[lower_guid].append(uefi_rule)

    if not rules_guids.keys():
        print(
            f"{error_prefix} None of the rules specify volume_guids (use scan command)"
        )
        return False

    with open(image_path, "rb") as f:
        firmware_data = f.read()

    extractor = UefiExtractor(firmware_data, list(rules_guids.keys()))
    extractor.extract_all(ignore_guid=False)

    if not len(extractor.binaries):
        print("No modules were found for scanning")
        return False

    for binary in extractor.binaries:
        fpath = os.path.join(tempfile.gettempdir(), f"{binary.name}{binary.ext}")
        with open(fpath, "wb") as f:
            f.write(binary.content)

        uefi_analyzer = UefiAnalyzer(image_path=fpath)
        scanner = UefiScanner(uefi_analyzer, rules_guids[binary.guid])

        for result in scanner.results:
            msg = threat if result.res else no_threat
            print(
                f"{prefix} {result.rule.name} (variant: {result.variant_label}) {msg} ({binary.name})"
            )

        uefi_analyzer.close()
        os.remove(fpath)

    return True


@click.command()
@click.argument(
    "image_path", type=click.Path(exists=True, dir_okay=False, file_okay=True)
)
@click.argument("extract_path", type=click.Path(dir_okay=True, file_okay=False))
def extract(image_path: str, extract_path: str) -> bool:
    """Extract all modules from UEFI firmware image."""

    if not os.path.isdir(extract_path):
        os.mkdir(extract_path)

    with open(image_path, "rb") as f:
        firmware_data = f.read()

    extractor = UefiExtractor(firmware_data, list())
    extractor.extract_all(ignore_guid=True)

    if not len(extractor.binaries):
        click.echo("No modules found", err=True)
        return False

    for binary in extractor.binaries:
        if not binary.guid or not len(binary.content):
            continue
        fpath = os.path.join(extract_path, f"{binary.guid}-{binary.name}{binary.ext}")
        with open(fpath, "wb") as f:
            f.write(binary.content)

        click.echo(f"{binary.guid} -> {fpath}")

    return True


cli.add_command(analyze_module)
cli.add_command(scan_module)
cli.add_command(scan_firmware)
cli.add_command(extract)

if __name__ == "__main__":
    cli()



from _typeshed import Incomplete
from fwhunt_scan.uefi_analyzer import NvramVariable as NvramVariable, UefiAnalyzer as UefiAnalyzer, UefiGuid as UefiGuid, UefiProtocol as UefiProtocol, UefiService as UefiService
from typing import Any, Dict, List, Optional

class CodePattern:
    pattern: Incomplete
    place: Incomplete
    unspecified: Incomplete
    def __init__(self, pattern: str, place: Optional[str]) -> None: ...
    @property
    def __dict__(self): ...

class UefiRuleVariant:
    def __init__(self, rule_content: Dict[str, Any]) -> None: ...
    @property
    def code(self) -> Dict[str, List[CodePattern]]: ...
    @property
    def strings(self) -> Dict[str, List[str]]: ...
    @property
    def wide_strings(self) -> Dict[str, List[Dict[str, str]]]: ...
    @property
    def hex_strings(self) -> Dict[str, List[str]]: ...
    @property
    def nvram_vars(self) -> Dict[str, List[NvramVariable]]: ...
    @property
    def protocols(self) -> Dict[str, List[UefiProtocol]]: ...
    @property
    def ppi_list(self) -> Dict[str, List[UefiProtocol]]: ...
    @property
    def guids(self) -> Dict[str, List[UefiGuid]]: ...

class UefiRule:
    def __init__(self, rule_path: Optional[str] = ..., rule_content: Optional[str] = ...) -> None: ...
    @property
    def author(self) -> Optional[str]: ...
    @property
    def name(self) -> Optional[str]: ...
    @property
    def version(self) -> Optional[str]: ...
    @property
    def namespace(self) -> Optional[str]: ...
    @property
    def license(self) -> Optional[str]: ...
    @property
    def cve_number(self) -> Optional[str]: ...
    @property
    def volume_guids(self) -> Optional[List[str]]: ...
    @property
    def vendor_id(self) -> Optional[str]: ...
    @property
    def cvss_score(self) -> Optional[str]: ...
    @property
    def advisory(self) -> Optional[str]: ...
    @property
    def description(self) -> Optional[str]: ...
    @property
    def url(self) -> Optional[str]: ...
    @property
    def target(self) -> Optional[str]: ...
    @property
    def variants(self) -> Dict[str, UefiRuleVariant]: ...

class UefiScannerError(Exception):
    value: Incomplete
    def __init__(self, value: str) -> None: ...

class UefiScannerRes:
    rule: Incomplete
    variant_label: Incomplete
    res: Incomplete
    def __init__(self, rule: UefiRule, variant_label: str, res: bool) -> None: ...

class UefiScanner:
    PROTOCOL: int
    PPI: int
    def __init__(self, uefi_analyzer: UefiAnalyzer, uefi_rules: List[UefiRule]) -> None: ...
    @property
    def results(self) -> List[UefiScannerRes]: ...



# SPDX-License-Identifier: GPL-3.0+
#
# pylint: disable=too-many-nested-blocks,invalid-name
# pylint: disable=too-few-public-methods,too-many-arguments,too-many-instance-attributes

"""
Tools for analyzing UEFI firmware using radare2/rizin
"""

import json
import string
import sys
import uuid
from types import TracebackType
from typing import Any, Dict, List, Optional, Type

import rzpipe

import fwhunt_scan.uefi_smm as uefi_smm
from fwhunt_scan.uefi_protocols import GUID_FROM_BYTES
from fwhunt_scan.uefi_tables import (
    BS_PROTOCOLS_INFO_64_BIT,
    EFI_BOOT_SERVICES_64_BIT,
    EFI_PEI_SERVICES_32_BIT,
    EFI_RUNTIME_SERVICES_64_BIT,
    OFFSET_TO_SERVICE,
)
from fwhunt_scan.uefi_te import TerseExecutableError, TerseExecutableParser
from fwhunt_scan.uefi_types import (
    ChildSwSmiHandler,
    NvramVariable,
    SmiHandler,
    UefiGuid,
    UefiProtocol,
    UefiProtocolGuid,
    UefiService,
)
from fwhunt_scan.uefi_utils import get_current_insn_index, get_int

if sys.version_info.major == 3 and sys.version_info.minor >= 8:
    from multiprocessing import shared_memory
if sys.version_info.major == 3 and (
    sys.version_info.minor >= 6 and sys.version_info.minor < 8
):
    import shared_memory  # type: ignore # noqa: F811


class UefiAnalyzerError(Exception):
    """Generic TE format error exception."""

    def __init__(self, value: str) -> None:
        self.value = value

    def __str__(self):
        return repr(self.value)


class UefiAnalyzer:
    """Helper object to analyze the EFI binary and provide properties"""

    def __init__(
        self,
        image_path: Optional[str] = None,
        blob: Optional[bytes] = None,
        rizinhome: Optional[str] = None,
    ):
        """UEFI analyzer initialization"""

        self._rz: rzpipe.open = None
        self._shm: Optional[shared_memory.SharedMemory] = None
        self._te: Optional[TerseExecutableParser] = None

        # init rizin
        if image_path:
            self._rz = rzpipe.open(
                filename=image_path, flags=["-2"], rizin_home=rizinhome
            )
            # analyze image
            with open(image_path, "rb") as f:
                if f.read(2) in [b"MZ", b"VZ"]:
                    self._rz.cmd("aaaa")
            try:
                self._te = TerseExecutableParser(image_path=image_path)
            except TerseExecutableError:
                self._te = None

        if blob and sys.platform in ["linux"]:
            blob_size = len(blob)
            self._shm = shared_memory.SharedMemory(create=True, size=blob_size)
            self._shm.buf[:] = blob[:]
            self._rz = rzpipe.open(
                filename=f"shm://{self._shm.name}/{blob_size:#d}",
                flags=["-2"],
                rizin_home=rizinhome,
            )
            if blob[:2] in [b"MZ", b"VZ"]:
                self._rz.cmd("aaaa")
            try:
                self._te = TerseExecutableParser(blob=blob)
            except TerseExecutableError:
                self._te = None

        if self._rz is None:
            raise UefiAnalyzerError(
                "Failed to initialize radare2/rizin analyzing engine"
            )

        # private cache
        self._bs_list_g_bs: Optional[List[UefiService]] = None
        self._bs_prot: Optional[List[UefiService]] = None
        self._rt_list: Optional[List[UefiService]] = None
        self._pei_services: Optional[List[UefiService]] = None
        self._ppi_list: Optional[List[UefiProtocol]] = None
        self._protocols: Optional[List[UefiProtocol]] = None
        self._protocol_guids: Optional[List[UefiProtocolGuid]] = None
        self._nvram_vars: Optional[List[NvramVariable]] = None
        self._info: Optional[Dict[Any, Any]] = None
        self._strings: Optional[List[Any]] = None
        self._sections: Optional[List[Any]] = None
        self._functions: Optional[List[Any]] = None
        self._insns: Optional[List[Any]] = None
        self._g_bs: Optional[List[int]] = None
        self._g_rt: Optional[List[int]] = None
        self._smst_list: Optional[List[int]] = None

        self._rt_filter_list: List[int] = list()
        self._bs_filter_list: List[int] = list()
        self._pei_filter_list: List[int] = list()

        # SMI handlers addresses
        self._smi_handlers: Optional[List[SmiHandler]] = None
        self._child_swsmi_handlers: Optional[List[ChildSwSmiHandler]] = None

    def __enter__(self):
        return self

    def _section_paddr(self, section_name: str) -> int:
        for section in self.sections:
            if section["name"] == section_name:
                return section["paddr"]
        return 0

    def _correct_addr(self, addr: int) -> int:
        if not self._te:
            return addr
        offset = (
            self._te.image_base + self._te.base_of_code - self._section_paddr(".text")
        )
        return addr + offset

    def _wrong_addr(self, addr: int) -> int:
        if not self._te:
            return addr
        offset = (
            self._te.image_base + self._te.base_of_code - self._section_paddr(".text")
        )
        return addr - offset

    def _pei_service_args_num(self, reg: str, addr: int) -> int:
        """Get number of arguments for specified PEI service call"""

        args_num = 0
        self._rz.cmd("s {:#x}".format(addr))
        res = self._rz.cmdj("pdbj")
        if not res:
            return 0

        # find current call in block
        for index in range(len(res)):
            if res[index]["offset"] == addr:
                break

        # get number of arguments
        for i in range(index, -1, -1):
            esil = res[i]["esil"].split(",")
            if len(esil) < 4:  # fix IndexError
                continue
            if esil[-3] == "4" and esil[-2] == "esp" and esil[-1] == "-=":
                args_num += 1
            if (
                len(esil) == 4
                and esil[-1] == "="
                and esil[-2] == reg
                and esil[-3] == "[4]"
            ):
                return args_num
        return 0

    @property
    def info(self) -> Dict[Any, Any]:
        """Get common image properties (parsed header)"""

        if self._info is None:
            self._info = self._rz.cmdj("ij")
        return self._info

    @property
    def strings(self) -> List[Any]:
        """Get common image properties (strings)"""

        if self._strings is None:
            self._strings = self._rz.cmdj("izzj") or []
        return self._strings

    @property
    def sections(self) -> List[Any]:
        """Get common image properties (sections)"""

        if self._sections is None:
            self._sections = self._rz.cmdj("iSj") or []
        return self._sections

    @property
    def functions(self) -> List[Any]:
        """Get common image properties (functions)"""

        if self._functions is None:
            self._functions = self._rz.cmdj("aflj") or []
        return self._functions

    def _get_insns(self) -> List[Any]:
        insns = list()
        for function in self.functions:
            self._rz.cmd("s {:#x}".format(function["offset"]))
            insns += self._rz.cmdj("pDj {:#x}".format(function["size"]))
        return insns

    @property
    def insns(self) -> List[Any]:
        """Get instructions"""

        if self._insns is None:
            self._insns = self._get_insns()
        return self._insns

    def _get_bs_64bit(self) -> List[int]:
        res: List[int] = list()

        for func in self.functions:
            func_addr = func["offset"]
            func_insns = self._rz.cmdj("pdfj @ {:#x}".format(func_addr))
            g_bs_reg = None
            for insn in func_insns["ops"]:
                if "esil" in insn:
                    esil = insn["esil"].split(",")
                    if (
                        get_int(esil[0]) == 0x60
                        and (esil[2] == "+")
                        and (esil[3] == "[8]")
                        and (esil[-1] == "=")
                    ):
                        g_bs_reg = esil[-2]
                    if not g_bs_reg:
                        continue
                    if (
                        (esil[0] == g_bs_reg)
                        and (esil[-1] == "=[8]")
                        and ("ptr" in insn)
                    ):
                        res.append(insn["ptr"])
                        break
        return res

    @property
    def g_bs(self) -> List[int]:
        """Find BootServices table global address"""

        if self._g_bs is None:
            self._g_bs = self._get_bs_64bit()
        return self._g_bs

    def _get_rt_64bit(self) -> List[int]:
        res: List[int] = list()

        for func in self.functions:
            func_addr = func["offset"]
            func_insns = self._rz.cmdj("pdfj @ {:#x}".format(func_addr))
            g_rt_reg = None
            for insn in func_insns["ops"]:
                if "esil" in insn:
                    esil = insn["esil"].split(",")
                    if (
                        get_int(esil[0]) == 0x58
                        and (esil[2] == "+")
                        and (esil[3] == "[8]")
                        and (esil[-1] == "=")
                    ):
                        g_rt_reg = esil[-2]
                    if not g_rt_reg:
                        continue
                    if (
                        (esil[0] == g_rt_reg)
                        and (esil[-1] == "=[8]")
                        and ("ptr" in insn)
                    ):
                        res.append(insn["ptr"])
                        break
        return res

    @property
    def g_rt(self) -> List[int]:
        """Find RuntimeServices table global address"""

        if self._g_rt is None:
            self._g_rt = self._get_rt_64bit()
        return self._g_rt

    def _get_boot_services_bs_64bit(self) -> List[UefiService]:
        bs_list: List[UefiService] = list()
        addrs: List[int] = list()

        if not len(self.g_bs):
            return bs_list

        for insn in self.insns:
            # find "mov rax, qword [g_bs]" instruction
            found = False
            if "esil" not in insn:
                continue
            esil = insn["esil"].split(",")
            if (
                (insn["type"] == "mov")
                and (esil[-1] == "=")
                and (esil[-3] == "[8]")
                and (esil[-4] == "+")
            ):
                if ("ptr" in insn) and (insn["ptr"] in self.g_bs):
                    found = True
            if not found:
                continue

            # if current instriction is "mov rax, qword [g_bs]"
            index = self.insns.index(insn)
            for g_bs_area_insn in self.insns[index : index + 0x10]:
                if "esil" not in g_bs_area_insn.keys():
                    continue
                g_bs_area_esil = g_bs_area_insn["esil"].split(",")
                if not (
                    (g_bs_area_insn["type"] in ["ucall", "ircall", "ujmp", "irjmp"])
                    and (g_bs_area_esil[1] == "rax")
                    and (g_bs_area_esil[2] == "+")
                    and (g_bs_area_esil[3] == "[8]")
                    and (g_bs_area_esil[-1] == "=")
                ):
                    continue
                if "ptr" not in g_bs_area_insn:
                    continue
                service_offset = g_bs_area_insn["ptr"]
                if service_offset in EFI_BOOT_SERVICES_64_BIT:
                    if g_bs_area_insn["offset"] in addrs:
                        break
                    if g_bs_area_insn["offset"] not in self._bs_filter_list:
                        bs_list.append(
                            UefiService(
                                address=g_bs_area_insn["offset"],
                                name=EFI_BOOT_SERVICES_64_BIT[service_offset],
                            )
                        )
                        self._bs_filter_list.append(g_bs_area_insn["offset"])
                    break

        return bs_list

    def _get_boot_services_prot_64bit(self) -> List[UefiService]:
        bs_prot: List[UefiService] = list()

        for insn in self.insns:
            if "esil" not in insn:
                continue
            esil = insn["esil"].split(",")
            if not (
                (insn["type"] in ["ucall", "ircall", "ujmp", "irjmp"])
                and (esil[1] == "rax")
                and (esil[2] == "+")
                and (esil[3] == "[8]")
            ):
                continue
            if "ptr" not in insn:
                continue
            service_offset = insn["ptr"]
            if service_offset in OFFSET_TO_SERVICE:
                name = OFFSET_TO_SERVICE[service_offset]
                # found boot service that work with protocol
                if insn["offset"] not in self._bs_filter_list:
                    bs_prot.append(UefiService(address=insn["offset"], name=name))
                    self._bs_filter_list.append(insn["offset"])
        return bs_prot

    @property
    def boot_services(self) -> List[UefiService]:
        """Find boot services using g_bs"""

        if self._bs_prot is None:
            self._bs_prot = self._get_boot_services_prot_64bit()
        if self._bs_list_g_bs is None:
            self._bs_list_g_bs = self._get_boot_services_bs_64bit()
        return self._bs_list_g_bs + self._bs_prot

    @property
    def boot_services_protocols(self) -> List[Any]:
        """Find boot service that work with protocols"""

        if self._bs_prot is None:
            self._bs_prot = self._get_boot_services_prot_64bit()
        return self._bs_prot

    def _get_runtime_services_64bit(self) -> List[UefiService]:
        rt_list: List[UefiService] = list()

        if not len(self.g_rt):
            return rt_list

        for insn in self.insns:
            # find "mov rax, qword [g_rt]" instruction
            found = False
            if "esil" not in insn:
                continue
            esil = insn["esil"].split(",")
            if (
                (insn["type"] == "mov")
                and (esil[-1] == "=")
                and (esil[-3] == "[8]")
                and (esil[-4] == "+")
            ):
                if ("ptr" in insn) and (insn["ptr"] in self.g_rt):
                    found = True
            if not found:
                continue

            index = self.insns.index(insn)
            for g_rt_area_insn in self.insns[index : index + 0x10]:
                g_rt_area_esil = g_rt_area_insn["esil"].split(",")
                if not (
                    (g_rt_area_insn["type"] in ["ucall", "ircall", "ujmp", "irjmp"])
                    and (g_rt_area_esil[1] == "rax")
                    and (g_rt_area_esil[2] == "+")
                    and (g_rt_area_esil[3] == "[8]")
                    and (g_rt_area_esil[-1] == "=")
                ):
                    continue
                if "ptr" not in g_rt_area_insn:
                    continue
                service_offset = g_rt_area_insn["ptr"]
                if service_offset in EFI_RUNTIME_SERVICES_64_BIT:
                    if g_rt_area_insn["offset"] not in self._rt_filter_list:
                        rt_list.append(
                            UefiService(
                                address=g_rt_area_insn["offset"],
                                name=EFI_RUNTIME_SERVICES_64_BIT[service_offset],
                            )
                        )
                        self._rt_filter_list.append(g_rt_area_insn["offset"])
                    break
        return rt_list

    @property
    def runtime_services(self) -> List[UefiService]:
        """Find all runtime services"""

        if self._rt_list is None:
            self._rt_list = self._get_runtime_services_64bit()
        return self._rt_list

    def _get_protocols_64bit(self) -> List[UefiProtocol]:
        protocols = list()
        for bs in self.boot_services_protocols:
            block_insns = self._rz.cmd("pdbj @ {:#x}".format(bs.address))
            try:
                block_insns = json.loads(block_insns)
            except (ValueError, KeyError, TypeError):
                continue
            for insn in block_insns:
                if "esil" in insn:
                    esil = insn["esil"].split(",")
                    if not (
                        (insn["type"] == "lea")
                        and (esil[-1] == "=")
                        and (esil[-2] == BS_PROTOCOLS_INFO_64_BIT[bs.name]["reg"])
                        and (esil[-3] == "+")
                    ):
                        continue
                    if "ptr" not in insn:
                        continue
                    p_guid_addr = insn["ptr"]
                    self._rz.cmd("s {:#x}".format(p_guid_addr))
                    p_guid_b = bytes(self._rz.cmdj("xj 16"))

                    # look up in known list
                    guid = GUID_FROM_BYTES.get(p_guid_b)
                    if not guid:
                        guid = UefiGuid(
                            value=str(uuid.UUID(bytes_le=p_guid_b)).upper(),
                            name="UNKNOWN_PROTOCOL",
                        )

                    protocols.append(
                        UefiProtocol(
                            name=guid.name,
                            value=guid.value,
                            guid_address=p_guid_addr,
                            address=insn["offset"],
                            service=bs.name,
                        )
                    )
        return protocols

    @property
    def protocols(self) -> List[UefiProtocol]:
        """Find proprietary protocols"""

        if self._protocols is None:
            self._protocols = self._get_protocols_64bit()
        return self._protocols

    def _get_protocol_guids(self) -> List[UefiProtocolGuid]:
        protocol_guids = list()
        target_sections = [".data"]
        for section in self.sections:
            if section["name"] in target_sections:
                self._rz.cmd("s {:#x}".format(section["vaddr"]))
                section_data = bytes(self._rz.cmdj("xj {:#d}".format(section["vsize"])))

                # find guids in section data:
                for i in range(len(section_data) - 15):
                    chunk = section_data[i : i + 16]
                    guid = GUID_FROM_BYTES.get(chunk)
                    if not guid:
                        continue
                    if guid.value in ["00000000-0000-0000-0000000000000000"]:
                        continue
                    protocol_guids.append(
                        UefiProtocolGuid(
                            address=section["vaddr"] + i,
                            name=guid.name,
                            value=guid.value,
                        )
                    )
        return protocol_guids

    @property
    def protocol_guids(self) -> List[UefiProtocolGuid]:
        """Find protocols GUIDs"""

        if self._protocol_guids is None:
            self._protocol_guids = self._get_protocol_guids()
        return self._protocol_guids

    @staticmethod
    def _name_is_valid(name: str) -> bool:
        for c in name:
            if c not in string.printable:
                return False
        return True

    def r2_get_nvram_vars_64bit(self) -> List[NvramVariable]:
        # to fix FPs need to apply filtering
        # of the extracted Runtime Services (from self.runtime_services)

        nvram_vars = list()
        for service in self.runtime_services:
            if service.name not in ["GetVariable", "SetVariable"]:
                continue

            # disassemble current function
            func_insns = self._rz.cmdj("pdfj @ {:#x}".format(service.address))
            if "ops" not in func_insns:
                continue

            insns = func_insns["ops"]
            index = get_current_insn_index(insns, service.address)
            if index is None:
                continue

            name: Optional[str] = None
            name_addr_reg: Optional[str] = None
            p_guid_b: Optional[bytes] = None
            stack_guid: bool = False
            for i in range(index - 1, -1, -1):
                insn = insns[i]
                if name is not None and p_guid_b is not None:
                    break

                if "esil" not in insn:
                    continue
                esil = insn["esil"].split(",")

                # handle case when we have:
                # NAME_ADDR,rip,+,REG,=
                # REG,rcx,=
                if (
                    name is None
                    and name_addr_reg is None
                    and len(esil) == 3
                    and (esil[-1] == "=" and esil[-2] == "rcx")
                ):
                    name_addr_reg = esil[0]

                ptr = insn.get("ptr", None)
                if ptr is None:
                    continue

                if (
                    name is None
                    and name_addr_reg
                    and len(esil) == 5
                    and esil[-1] == "="
                    and esil[-2] == name_addr_reg
                    and esil[-3] == "+"
                    and esil[-4] == "rip"
                ):
                    tmp_name = self._rz.cmd("psw @ {:#x}".format(ptr))[:-1]
                    if UefiAnalyzer._name_is_valid(tmp_name):
                        name = tmp_name

                if (
                    name is None
                    and len(esil) == 5
                    and (esil[-1] == "=")
                    and (esil[-2] == "rcx")
                    and (esil[-3] == "+")
                    and (esil[-4] == "rip")
                ):
                    tmp_name = self._rz.cmd("psw @ {:#x}".format(ptr))[:-1]
                    if UefiAnalyzer._name_is_valid(tmp_name):
                        name = tmp_name

                if (
                    p_guid_b is None
                    and len(esil) == 5
                    and (esil[-1] == "=")
                    and (esil[-2] == "rdx")
                    and (esil[-4] in ["rsp", "rbp"])
                ):
                    stack_guid = True

                if (
                    not stack_guid
                    and p_guid_b is None
                    and len(esil) == 5
                    and (esil[-1] == "=")
                    and (esil[-2] == "rdx")
                    and (esil[-3] == "+")
                    and (esil[-4] == "rip")
                ):
                    p_guid_b = bytes(self._rz.cmdj("xj 16 @ {:#x}".format(ptr)))

            if not name:
                name = "Unknown"

            if p_guid_b is not None:
                guid = str(uuid.UUID(bytes_le=p_guid_b)).upper()
            else:
                guid = "Unknown"

            nvram_vars.append(NvramVariable(name=name, guid=guid, service=service))

        return nvram_vars

    @property
    def nvram_vars(self) -> List[NvramVariable]:
        """Find NVRAM variables passed to GetVariable and SetVariable services"""

        if self._nvram_vars is None:
            self._nvram_vars = self.r2_get_nvram_vars_64bit()
        return self._nvram_vars

    def _get_pei_services(self) -> List[UefiService]:
        pei_list: List[UefiService] = list()
        for func in self.functions:
            func_addr = func["offset"]
            func_insns = self._rz.cmdj("pdfj @ {:#x}".format(func_addr))
            for insn in func_insns["ops"]:
                if "esil" not in insn:
                    continue
                esil = insn["esil"].split(",")
                if esil[-1] == "=" and esil[-2] == "eip":
                    offset = get_int(esil[0])
                    if offset is None:
                        continue
                    if offset not in EFI_PEI_SERVICES_32_BIT.keys():
                        continue
                    reg = esil[1]
                    service: Dict[str, Any] = EFI_PEI_SERVICES_32_BIT[offset]

                    # found potential pei service, compare number of arguments
                    arg_num: Any = self._pei_service_args_num(reg, insn["offset"])
                    if arg_num < service["arg_num"]:
                        continue

                    # wrong addresses in r2 in case of TE
                    if insn["offset"] not in self._pei_filter_list:
                        pei_list.append(
                            UefiService(address=insn["offset"], name=service["name"])
                        )
                        self._pei_filter_list.append(insn["offset"])

        return pei_list

    @property
    def pei_services(self) -> List[UefiService]:
        """Find all PEI services"""

        if self._pei_services is None:
            self._pei_services = self._get_pei_services()
        return self._pei_services

    def _get_ppi_list(self) -> List[UefiProtocol]:
        ppi_list: List[UefiProtocol] = list()
        for pei_service in self.pei_services:
            if pei_service.name != "LocatePpi":
                continue
            block_insns = self._rz.cmdj("pdj -16 @ {:#x}".format(pei_service.address))
            for index in range(len(block_insns) - 1, -1, -1):
                esil = block_insns[index]["esil"].split(",")
                if not (esil[-1] == "-=" and esil[-2] == "esp" and esil[-3] == "4"):
                    continue
                if "ptr" not in block_insns[index]:
                    continue
                current_block = block_insns[index]
                p_guid_addr = current_block["ptr"]
                baddr = self.info["bin"]["baddr"]
                if p_guid_addr < baddr:
                    continue
                # wrong addresses in r2 in case of TE
                p_guid_addr = self._wrong_addr(p_guid_addr)
                self._rz.cmd("s {:#x}".format(p_guid_addr))
                p_guid_b = bytes(self._rz.cmdj("xj 16"))
                # look up in known list
                guid = GUID_FROM_BYTES.get(p_guid_b)
                if not guid:
                    guid = UefiGuid(
                        value=str(uuid.UUID(bytes_le=p_guid_b)).upper(),
                        name="proprietary_ppi",
                    )
                ppi = UefiProtocol(
                    name=guid.name,
                    value=guid.value,
                    guid_address=p_guid_addr,
                    address=block_insns[index]["offset"],
                    service=pei_service.name,
                )
                if ppi not in ppi_list:
                    ppi_list.append(ppi)

        return ppi_list

    @property
    def ppi_list(self) -> List[UefiProtocol]:
        """Find all PPIs"""

        if self._ppi_list is None:
            self._ppi_list = self._get_ppi_list()
        return self._ppi_list

    @property
    def smi_handlers(self) -> List[SmiHandler]:
        """Find software SMI handlers"""

        if self._smi_handlers is None:
            self._smi_handlers = uefi_smm.get_smi_handlers(self._rz)
        return self._smi_handlers

    @property
    def child_swsmi_handlers(self) -> List[ChildSwSmiHandler]:
        """Find child software SMI handlers"""

        if self._child_swsmi_handlers is None:
            self._child_swsmi_handlers = uefi_smm.get_child_sw_smi_handlers(
                self._rz, self.smst_list
            )
        return self._child_swsmi_handlers

    @property
    def smst_list(self) -> List[int]:
        """Find list of SMST"""

        if self._smst_list is None:
            self._smst_list = uefi_smm.get_smst_list(self._rz)
        return self._smst_list

    def get_summary(self) -> Dict[str, Any]:
        """Collect all the information in a JSON object"""

        summary = dict()

        for key in self.info:
            summary[key] = self.info[key]

        if "bin" not in summary:
            return summary

        if not (
            self.info["bin"]["class"].startswith("PE")
            or self.info["bin"]["class"].startswith("TE")
        ):
            return summary

        if self.info["bin"]["arch"] == "x86" and self.info["bin"]["bits"] == 32:
            summary["pei_list"] = [x.__dict__ for x in self.pei_services]
            summary["ppi_list"] = [x.__dict__ for x in self.ppi_list]
            summary["nvram_vars"] = [x.__dict__ for x in self.nvram_vars]

        elif self.info["bin"]["arch"] == "x86" and self.info["bin"]["bits"] == 64:
            summary["g_bs"] = self.g_bs
            summary["g_rt"] = self.g_rt
            summary["g_smst"] = self.smst_list
            summary["bs_list"] = [x.__dict__ for x in self.boot_services]
            summary["rt_list"] = [x.__dict__ for x in self.runtime_services]
            summary["protocols"] = [x.__dict__ for x in self.protocols]
            summary["nvram_vars"] = [x.__dict__ for x in self.nvram_vars]
            if len(self.smi_handlers) > 0:
                summary["smi_handlers"] = [x.__dict__ for x in self.smi_handlers]
            if len(self.child_swsmi_handlers) > 0:
                summary["child_swsmi_handlers"] = [
                    x.__dict__ for x in self.child_swsmi_handlers
                ]

        summary["p_guids"] = [x.__dict__ for x in self.protocol_guids]

        return summary

    def get_protocols_info(self) -> Dict[str, Any]:
        """Collect all the information in a JSON object"""

        summary = dict()
        for key in self.info:
            summary[key] = self.info[key]
        summary["g_bs"] = self.g_bs
        summary["bs_list"] = [x.__dict__ for x in self.boot_services]
        summary["protocols"] = [x.__dict__ for x in self.protocols]
        return summary

    def close(self) -> None:
        """Quits the r2 instance, releasing resources"""

        self._rz.quit()
        if self._shm is not None:
            self._shm.close()
            self._shm.unlink()

    def __exit__(
        self,
        exc_type: Optional[Type[BaseException]],
        exc_value: Optional[BaseException],
        traceback: Optional[TracebackType],
    ) -> None:
        self.close()



# SPDX-License-Identifier: GPL-3.0+

import uuid
from enum import Enum
from typing import Optional


class SmiKind(Enum):
    CHILD_SW_SMI = 0
    SW_SMI = 1
    USB_SMI = 2
    SX_SMI = 3
    IO_TRAP_SMI = 4
    GPI_SMI = 5
    TCO_SMI = 6
    STANDBY_BUTTON_SMI = 7
    PERIODIC_TIMER_SMI = 8
    POWER_BUTTON_SMI = 9
    ICHN_SMI = 10
    PCH_TCO_SMI = 11
    PCH_PCIE_SMI = 12
    PCH_ACPI_SMI = 13
    PCH_GPIO_UNLOCK_SMI = 14
    PCH_SMI = 15
    PCH_ESPI_SMI = 16
    ACPI_EN_SMI = 17
    ACPI_DIS_SMI = 18


class UefiService:
    """A UEFI service"""

    def __init__(self, name: str, address: int) -> None:
        self.name: str = name
        self.address: int = address

    @property
    def __dict__(self):
        val = dict()
        if self.name:
            val["name"] = self.name
        if self.address:
            val["address"] = self.address
        return val


class UefiGuid:
    """A UEFI GUID"""

    def __init__(self, value: str, name: str) -> None:
        self.value: str = value
        self.name: str = name
        self._bytes: bytes = b""

    @property
    def bytes(self) -> bytes:
        """Convert guid structure to array of bytes"""
        if not self._bytes:
            self._bytes = uuid.UUID(self.value).bytes_le
        return self._bytes

    @property
    def __dict__(self):
        return {"value": self.value, "name": self.name}

    def __str__(self):
        return "{} ({})".format(self.value, self.name)


class UefiProtocol(UefiGuid):
    """A UEFI protocol"""

    def __init__(
        self, name: str, address: int, value: str, guid_address: int, service: str
    ) -> None:
        super().__init__(name=name, value=value)
        self.address: int = address
        self.guid_address: int = guid_address
        self.service: str = service

    @property
    def __dict__(self):
        val = super().__dict__
        if self.address:
            val["address"] = self.address
        if self.guid_address:
            val["guid_address"] = self.guid_address
        if self.service:
            val["service"] = self.service
        return val


class UefiProtocolGuid(UefiGuid):
    """A UEFI protocol GUID"""

    def __init__(self, name: str, address: int, value: str) -> None:
        super().__init__(name=name, value=value)
        self.address: int = address

    @property
    def __dict__(self):
        val = super().__dict__
        if self.address:
            val["address"] = self.address
        return val


class NvramVariable:
    """A UEFI NVRAM variable"""

    def __init__(self, name: str, guid: str, service: UefiService) -> None:
        self.name: str = name
        self.guid: str = guid
        self.service: UefiService = service

    @property
    def __dict__(self):
        val = dict()
        if self.name:
            val["name"] = self.name
        if self.guid:
            val["guid"] = self.guid
        if self.service:
            val["service"] = {
                "name": self.service.name,
                "address": self.service.address,
            }
        return val


class SmiHandler:
    """SMI handler basic class"""

    def __init__(self, address: int, kind: SmiKind) -> None:
        self.address = address
        self.kind = kind
        self._place: Optional[str] = None

    def _get_place(self):
        return f"{self.kind.name.lower()}_handlers"

    @property
    def place(self):
        if self._place is None:
            self._place = self._get_place()
        return self._place

    @property
    def __dict__(self):
        val = dict()
        if self.address:
            val["address"] = self.address
        if self.kind:
            val["kind"] = self.kind.name
        return val


class ChildSwSmiHandler(SmiHandler):
    """Child software SMI handler"""

    def __init__(self, handler_guid: Optional[str], address: int) -> None:
        super().__init__(address=address, kind=SmiKind.CHILD_SW_SMI)
        self.handler_guid = handler_guid

    @property
    def __dict__(self):
        val = dict()
        if self.address:
            val["address"] = self.address
        if self.handler_guid:
            val["handler_guid"] = self.handler_guid
        if self.kind:
            val["kind"] = self.kind.name
        return val



from .uefi_analyzer import UefiAnalyzer as UefiAnalyzer, UefiAnalyzerError as UefiAnalyzerError
from .uefi_extractor import UefiBinary as UefiBinary, UefiExtractor as UefiExtractor
from .uefi_scanner import UefiRule as UefiRule, UefiScanner as UefiScanner, UefiScannerError as UefiScannerError
from .uefi_te import TerseExecutableParser as TerseExecutableParser



from fwhunt_scan import UefiAnalyzer, UefiRule, UefiScanner

rule_path = "test/rules/variants.yml"
rule = UefiRule(rule_path=rule_path)
for variant in rule.variants:
    code = rule.variants[variant].code
    print(variant, code)



from _typeshed import Incomplete
from typing import Optional

class TerseExecutableError(Exception):
    value: Incomplete
    def __init__(self, value: str) -> None: ...

class TerseExecutableParser:
    def __init__(self, image_path: Optional[str] = ..., blob: Optional[bytes] = ...) -> None: ...
    @property
    def signature(self) -> str: ...
    @property
    def machine(self) -> int: ...
    @property
    def number_of_sections(self) -> int: ...
    @property
    def subsystem(self) -> int: ...
    @property
    def stripped_size(self) -> int: ...
    @property
    def address_of_entry_point(self) -> int: ...
    @property
    def base_of_code(self) -> int: ...
    @property
    def image_base(self) -> int: ...



from _typeshed import Incomplete
from enum import Enum
from typing import Optional

class SmiKind(Enum):
    CHILD_SW_SMI: int
    SW_SMI: int
    USB_SMI: int
    SX_SMI: int
    IO_TRAP_SMI: int
    GPI_SMI: int
    TCO_SMI: int
    STANDBY_BUTTON_SMI: int
    PERIODIC_TIMER_SMI: int
    POWER_BUTTON_SMI: int
    ICHN_SMI: int
    PCH_TCO_SMI: int
    PCH_PCIE_SMI: int
    PCH_ACPI_SMI: int
    PCH_GPIO_UNLOCK_SMI: int
    PCH_SMI: int
    PCH_ESPI_SMI: int
    ACPI_EN_SMI: int
    ACPI_DIS_SMI: int

class UefiService:
    name: Incomplete
    address: Incomplete
    def __init__(self, name: str, address: int) -> None: ...
    @property
    def __dict__(self): ...

class UefiGuid:
    value: Incomplete
    name: Incomplete
    def __init__(self, value: str, name: str) -> None: ...
    @property
    def bytes(self) -> bytes: ...
    @property
    def __dict__(self): ...

class UefiProtocol(UefiGuid):
    address: Incomplete
    guid_address: Incomplete
    service: Incomplete
    def __init__(self, name: str, address: int, value: str, guid_address: int, service: str) -> None: ...
    @property
    def __dict__(self): ...

class UefiProtocolGuid(UefiGuid):
    address: Incomplete
    def __init__(self, name: str, address: int, value: str) -> None: ...
    @property
    def __dict__(self): ...

class NvramVariable:
    name: Incomplete
    guid: Incomplete
    service: Incomplete
    def __init__(self, name: str, guid: str, service: UefiService) -> None: ...
    @property
    def __dict__(self): ...

class SmiHandler:
    address: Incomplete
    kind: Incomplete
    def __init__(self, address: int, kind: SmiKind) -> None: ...
    @property
    def place(self): ...
    @property
    def __dict__(self): ...

class ChildSwSmiHandler(SmiHandler):
    handler_guid: Incomplete
    def __init__(self, handler_guid: Optional[str], address: int) -> None: ...
    @property
    def __dict__(self): ...



# SPDX-License-Identifier: GPL-3.0+

import json
import uuid
from typing import Any, Dict, List, Optional

import rzpipe

from fwhunt_scan.uefi_protocols import UefiGuid
from fwhunt_scan.uefi_types import ChildSwSmiHandler, SmiHandler, SmiKind
from fwhunt_scan.uefi_utils import (
    get_current_insn_index,
    get_int,
    get_xrefs_to_data,
    get_xrefs_to_guids,
)


SMI_KINDS = {
    SmiKind.SW_SMI: [
        UefiGuid(
            "18A3C6DC-5EEA-48C8-A1C1B53389F98999",
            name="EFI_SMM_SW_DISPATCH2_PROTOCOL_GUID",
        ),
        UefiGuid(
            "E541B773-DD11-420C-B026DF993653F8BF",
            name="EFI_SMM_SW_DISPATCH_PROTOCOL_GUID",
        ),
    ],
    SmiKind.USB_SMI: [
        UefiGuid(
            "EE9B8D90-C5A6-40A2-BDE252558D33CCA1",
            name="EFI_SMM_USB_DISPATCH2_PROTOCOL_GUID",
        ),
        UefiGuid(
            "A05B6FFD-87AF-4E42-95C96228B63CF3F3",
            name="EFI_SMM_USB_DISPATCH_PROTOCOL_GUID",
        ),
    ],
    SmiKind.SX_SMI: [
        UefiGuid(
            "456D2859-A84B-4E47-A2EE3276D886997D",
            name="EFI_SMM_SX_DISPATCH2_PROTOCOL_GUID",
        ),
        UefiGuid(
            "14FC52BE-01DC-426C-91AEA23C3E220AE8",
            name="EFI_SMM_SX_DISPATCH_PROTOCOL_GUID",
        ),
    ],
    SmiKind.IO_TRAP_SMI: [
        UefiGuid(
            "58DC368D-7BFA-4E77-ABBC0E29418DF930",
            name="EFI_SMM_IO_TRAP_DISPATCH2_PROTOCOL_GUID",
        ),
        UefiGuid(
            "DB7F536B-EDE4-4714-A5C8E346EBAA201D",
            name="EFI_SMM_IO_TRAP_DISPATCH_PROTOCOL_GUID",
        ),
    ],
    SmiKind.GPI_SMI: [
        UefiGuid(
            "25566B03-B577-4CBF-958CED663EA24380",
            name="EFI_SMM_GPI_DISPATCH2_PROTOCOL_GUID",
        ),
        UefiGuid(
            "E0744B81-9513-49CD-8CEAE9245E7039DA",
            name="EFI_SMM_GPI_DISPATCH_PROTOCOL_GUID",
        ),
    ],
    SmiKind.STANDBY_BUTTON_SMI: [
        UefiGuid(
            "7300C4A1-43F2-4017-A51BC81A7F40585B",
            name="EFI_SMM_STANDBY_BUTTON_DISPATCH2_PROTOCOL_GUID",
        ),
        UefiGuid(
            "78965B98-B0BF-449E-8B22D2914E498A98",
            name="EFI_SMM_STANDBY_BUTTON_DISPATCH_PROTOCOL_GUID",
        ),
    ],
    SmiKind.PERIODIC_TIMER_SMI: [
        UefiGuid(
            "4CEC368E-8E8E-4D71-8BE1958C45FC8A53",
            name="EFI_SMM_PERIODIC_TIMER_DISPATCH2_PROTOCOL_GUID",
        ),
        UefiGuid(
            "9CCA03FC-4C9E-4A19-9B06ED7B479BDE55",
            name="EFI_SMM_PERIODIC_TIMER_DISPATCH_PROTOCOL_GUID",
        ),
    ],
    SmiKind.POWER_BUTTON_SMI: [
        UefiGuid(
            "1B1183FA-1823-46A7-88729C578755409D",
            name="EFI_SMM_POWER_BUTTON_DISPATCH2_PROTOCOL_GUID",
        ),
        UefiGuid(
            "B709EFA0-47A6-4B41-B93112ECE7A8EE56",
            name="EFI_SMM_POWER_BUTTON_DISPATCH_PROTOCOL_GUID",
        ),
    ],
    SmiKind.ICHN_SMI: [
        UefiGuid(
            "C50B323E-9075-4F2A-AC8ED2596A1085CC",
            name="EFI_SMM_ICHN_DISPATCH_PROTOCOL_GUID",
        ),
        UefiGuid(
            "ADF3A128-416D-4060-8DDF-30A1D7AAB699",
            name="EFI_SMM_ICHN_DISPATCH2_PROTOCOL_GUID",
        ),
    ],
    SmiKind.TCO_SMI: [
        UefiGuid(
            "0E2D6BB1-C624-446D-9982693CD181A607",
            name="EFI_SMM_TCO_DISPATCH_PROTOCOL_GUID",
        ),
    ],
    SmiKind.STANDBY_BUTTON_SMI: [
        UefiGuid(
            "7300C4A1-43F2-4017-A51BC81A7F40585B",
            name="EFI_SMM_STANDBY_BUTTON_DISPATCH2_PROTOCOL_GUID",
        ),
        UefiGuid(
            "78965B98-B0BF-449E-8B22D2914E498A98",
            name="EFI_SMM_STANDBY_BUTTON_DISPATCH_PROTOCOL_GUID",
        ),
    ],
    SmiKind.PERIODIC_TIMER_SMI: [
        UefiGuid(
            "4CEC368E-8E8E-4D71-8BE1958C45FC8A53",
            name="EFI_SMM_PERIODIC_TIMER_DISPATCH2_PROTOCOL_GUID",
        ),
        UefiGuid(
            "9CCA03FC-4C9E-4A19-9B06ED7B479BDE55",
            name="EFI_SMM_PERIODIC_TIMER_DISPATCH_PROTOCOL_GUID",
        ),
    ],
    SmiKind.POWER_BUTTON_SMI: [
        UefiGuid(
            "1B1183FA-1823-46A7-88729C578755409D",
            name="EFI_SMM_POWER_BUTTON_DISPATCH2_PROTOCOL_GUID",
        ),
        UefiGuid(
            "B709EFA0-47A6-4B41-B93112ECE7A8EE56",
            name="EFI_SMM_POWER_BUTTON_DISPATCH_PROTOCOL_GUID",
        ),
    ],
    SmiKind.PCH_TCO_SMI: [
        UefiGuid(
            "9E71D609-6D24-47FD-B572-6140F8D9C2A4",
            name="PCH_TCO_SMI_DISPATCH_PROTOCOL_GUID",
        ),
    ],
    SmiKind.PCH_PCIE_SMI: [
        UefiGuid(
            "3e7d2b56-3f47-42aa-8f6b-22f519818dab",
            name="PCH_PCIE_SMI_DISPATCH_PROTOCOL_GUID",
        ),
    ],
    SmiKind.PCH_ACPI_SMI: [
        UefiGuid(
            "d52bb262-f022-49ec-86d2-7a293a7a054b",
            name="PCH_ACPI_SMI_DISPATCH_PROTOCOL_GUID",
        ),
    ],
    SmiKind.PCH_GPIO_UNLOCK_SMI: [
        UefiGuid(
            "83339ef7-9392-4716-8d3a-d1fc67cd55db",
            name="PCH_GPIO_UNLOCK_SMI_DISPATCH_PROTOCOL_GUID",
        ),
    ],
    SmiKind.PCH_SMI: [
        UefiGuid(
            "e6a81bbf-873d-47fd-b6be-61b3e5720993",
            name="PCH_SMI_DISPATCH_PROTOCOL_GUID",
        ),
    ],
    SmiKind.PCH_ESPI_SMI: [
        UefiGuid(
            "b3c14ff3-bae8-456c-8631-27fe0ceb340c",
            name="PCH_ESPI_SMI_DISPATCH_PROTOCOL_GUID",
        ),
    ],
    SmiKind.ACPI_EN_SMI: [
        UefiGuid(
            "bd88ec68-ebe4-4f7b-935a-4f666642e75f",
            name="EFI_ACPI_EN_DISPATCH_PROTOCOL_GUID",
        ),
    ],
    SmiKind.ACPI_DIS_SMI: [
        UefiGuid(
            "9c939ba6-1fcc-46f6-b4e1-102dbe186567",
            name="EFI_ACPI_DIS_DISPATCH_PROTOCOL_GUID",
        ),
    ],
}


def get_interface_from_bb(insns: List[Dict[str, Any]], code_addr: int) -> Optional[int]:
    """Get the address of the interface
    (in the case of local variables, this will be the address on the stack)"""

    res = None

    index = get_current_insn_index(insns, code_addr)
    if index is None:
        return res

    # check all instructions from index to 0
    for i in range(index - 1, -1, -1):
        insn = insns[i]
        if "esil" not in insn:
            continue
        esil = insn["esil"].split(",")
        if not (esil[-1] == "=" and esil[-2] == "r8"):
            continue
        try:
            res = int(esil[0], 16)
            return res
        except ValueError:
            continue

    return res


def get_interface_global(insns: List[Dict[str, Any]], code_addr: int) -> Optional[int]:
    """Get the address of the interface
    (in the case of local variables, this will be the address on the stack)"""

    res = None

    index = get_current_insn_index(insns, code_addr)
    if index is None:
        return res

    # check all instructions from index to 0
    for i in range(index - 1, -1, -1):
        insn = insns[i]
        if "esil" not in insn:
            continue
        esil = insn["esil"].split(",")
        if not (esil[-1] == "=" and esil[-2] == "r8"):
            continue
        res = insn.get("ptr", None)
        if res is not None:
            return res
        res = get_int(esil[0])
        if res is not None:
            return res

    return res


def get_handler(insns: List[Dict[str, Any]], kind: SmiKind) -> Optional[SmiHandler]:
    address = None

    for insn in insns:
        if "esil" not in insn:
            continue
        esil = insn["esil"].split(",")

        if (
            len(esil) > 4
            and esil[-1] == "="
            and esil[-2] == "rdx"
            # skip esil[-3]
            # (esil[-3] == "+" or esil[-3] == "-")
            and esil[-4] == "rip"
        ):
            # handler address
            address = insn.get("ptr", None)

        # found `EfiSmmXXXDispatch2Protocol->Register()`
        if (
            esil == ["rax", "[8]", "rip", "8", "rsp", "-=", "rsp", "=[]", "rip", "="]
            and address is not None
        ):
            return SmiHandler(address=address, kind=kind)

    return None


def get_handlers(
    rz: rzpipe.open, code_addr: int, interface: int, kind: SmiKind
) -> List[SmiHandler]:
    res: List[SmiHandler] = list()
    res_addrs: List[int] = list()

    func = rz.cmdj(f"pdfj @ {code_addr:#x}")
    insns = func.get("ops", None)
    if insns is None:
        return res

    index = get_current_insn_index(insns, code_addr)
    if index is None:
        return res

    # check all instructions from index to end of function
    for i in range(index + 1, len(insns), 1):
        insn = insns[i]
        if "esil" not in insn:
            continue
        esil = insn["esil"].split(",")

        # find `mov rax, [rbp+EfiSmmSwDispatch2Protocol]` instruction
        if not (esil[-1] == "=" and esil[-2] == "rax"):
            continue

        value = get_int(esil[0])
        if value is None:
            continue

        if value == interface:
            offset = insn.get("offset", None)
            if offset is None:
                continue
            bb = rz.cmdj(f"pdbj @ {offset:#x}")
            handler = get_handler(bb, kind)
            if handler is not None:
                if handler.address not in res_addrs:
                    res.append(handler)
                    # prevent duplicates
                    res_addrs.append(handler.address)

    return res


def get_smst_bb(insns: List[Dict[str, Any]]) -> Optional[int]:
    res = None

    for insn in insns:
        if "esil" not in insn:
            continue
        esil = insn["esil"].split(",")

        # check esil insn ({offset},rip,+,rdx,=)
        if esil[-1] == "=" and esil[-2] == "rdx":
            res = insn.get("ptr", None)
            if res is not None:
                return res

    return res


def get_smst_func(rz: rzpipe.open, code_addr: int, interface: int) -> List[int]:
    res: List[int] = list()

    func = rz.cmdj(f"pdfj @ {code_addr:#x}")
    insns = func.get("ops", None)
    if insns is None:
        return res

    index = get_current_insn_index(insns, code_addr)
    if index is None:
        return res

    # check all instructions from index to end of function
    for i in range(index + 1, len(insns), 1):
        insn = insns[i]
        if "esil" not in insn:
            continue
        esil = insn["esil"].split(",")
        if esil[-1] == "=" and esil[-2] == "rax" and esil[-3] == "[8]":
            # check both for global variable and local variable
            if insn.get("ptr", None) == interface or get_int(esil[0]) == interface:
                offset = insn.get("offset", None)
                if offset is None:
                    continue
                bb = rz.cmdj(f"pdbj @ {offset:#x}")
                smst = get_smst_bb(bb)
                if smst is not None:
                    res.append(smst)
    return res


def get_smst_list(rz: rzpipe.open) -> List[int]:
    """Find SMST addresses"""

    res: List[int] = list()

    guids = [
        UefiGuid(
            "F4CCBFB7-F6E0-47FD-9DD410A8F150C191", name="EFI_SMM_BASE2_PROTOCOL_GUID"
        )
    ]
    code_addrs = get_xrefs_to_guids(rz, guids)
    for code_addr in code_addrs:
        bb = rz.cmdj(f"pdbj @ {code_addr:#x}")
        interface = get_interface_global(bb, code_addr)
        if interface is None:
            continue
        res += get_smst_func(rz, code_addr, interface)

    return res


def find_handler_register_service(insns: List[Dict[str, Any]]) -> Optional[int]:
    for insn in insns[::-1]:
        if "esil" not in insn:
            continue
        esil = insn["esil"].split(",")

        if esil[-7:] != ["8", "rsp", "-=", "rsp", "=[]", "rip", "="]:
            continue

        offset = get_int(esil[0])
        if offset == 0xE0:  # SmiHandlerRegister
            return insns.index(insn)

    return None


def get_child_sw_smi_handler_bb(
    rz: rzpipe.open, insns: List[Dict[str, Any]]
) -> Optional[ChildSwSmiHandler]:
    handler_address = None
    handler_guid = None

    end_index = find_handler_register_service(insns)
    if end_index is None:
        return None

    for insn in insns[:end_index][::-1]:
        if "esil" not in insn:
            continue
        esil = insn["esil"].split(",")

        # try to get handler address (Handler parameter)
        if esil[-1] == "=" and esil[-2] == "rcx":
            handler_address = insn.get("ptr", None)

        # try to get handler guid value (HandlerType parameter)
        if esil[-1] == "=" and esil[-2] == "rdx":
            guid_addr = insn.get("ptr", None)
            if guid_addr is not None:
                rz.cmd(f"s {guid_addr:#x}")
                guid_b = bytes(rz.cmdj("xj 16"))
                if len(guid_b) == 16:
                    handler_guid = str(uuid.UUID(bytes_le=guid_b)).upper()

        if handler_address is not None and handler_guid is not None:
            return ChildSwSmiHandler(address=handler_address, handler_guid=handler_guid)

    if handler_address is not None:  # handler_guid is Optional
        return ChildSwSmiHandler(address=handler_address, handler_guid=handler_guid)

    return None


def get_child_sw_smi_handlers(
    rz: rzpipe.open, smst_list: List[int]
) -> List[ChildSwSmiHandler]:
    res: List[ChildSwSmiHandler] = list()

    haddrs = list()  # addresses

    for smst in smst_list:
        code_addrs = get_xrefs_to_data(rz, smst)
        for addr in code_addrs:
            # analyze instructions and found gSmst->SmiHandlerRegister call
            result = rz.cmd(f"pdj 24 @ {addr:#x}")
            # prevent error messages to sys.stderr from rizin:
            # https://github.com/rizinorg/rz-pipe/blob/0f7ac66e6d679ebb03be26bf61a33f9ccf199f27/python/rzpipe/open_base.py#L261
            try:
                bb = json.loads(result)
            except (ValueError, KeyError, TypeError):
                continue
            handler = get_child_sw_smi_handler_bb(rz, bb)
            if handler is not None:
                if handler.address not in haddrs:
                    res.append(handler)
                    haddrs.append(handler.address)

    return res


def get_smi_handlers(rz: rzpipe.open) -> List[SmiHandler]:
    """Find Software SMI Handlers"""

    res: List[SmiHandler] = list()

    for kind in SMI_KINDS:
        code_addrs = get_xrefs_to_guids(rz, SMI_KINDS[kind])
        for code_addr in code_addrs:
            # get basic block information
            bb = rz.cmdj(f"pdbj @ {code_addr:#x}")
            interface = get_interface_from_bb(bb, code_addr)
            if interface is None:
                continue

            # need to check the use of this interface below code_addr
            res += get_handlers(rz, code_addr, interface, kind)

    return res



from _typeshed import Incomplete

EFI_BOOT_SERVICES_64_BIT: Incomplete
EFI_RUNTIME_SERVICES_64_BIT: Incomplete
EFI_SMM_SYSTEM_TABLE2_64_BIT: Incomplete
BS_PROTOCOLS: Incomplete
BS_PROTOCOLS_INFO_64_BIT: Incomplete
OFFSET_TO_SERVICE: Incomplete
EFI_PEI_SERVICES_32_BIT: Incomplete



import rzpipe
from _typeshed import Incomplete
from fwhunt_scan.uefi_protocols import UefiGuid as UefiGuid
from fwhunt_scan.uefi_types import ChildSwSmiHandler as ChildSwSmiHandler, SmiHandler as SmiHandler, SmiKind as SmiKind
from fwhunt_scan.uefi_utils import get_current_insn_index as get_current_insn_index, get_int as get_int, get_xrefs_to_data as get_xrefs_to_data, get_xrefs_to_guids as get_xrefs_to_guids
from typing import Any, Dict, List, Optional

SMI_KINDS: Incomplete

def get_interface_from_bb(insns: List[Dict[str, Any]], code_addr: int) -> Optional[int]: ...
def get_interface_global(insns: List[Dict[str, Any]], code_addr: int) -> Optional[int]: ...
def get_handler(insns: List[Dict[str, Any]], kind: SmiKind) -> Optional[SmiHandler]: ...
def get_handlers(rz: rzpipe.open, code_addr: int, interface: int, kind: SmiKind) -> List[SmiHandler]: ...
def get_smst_bb(insns: List[Dict[str, Any]]) -> Optional[int]: ...
def get_smst_func(rz: rzpipe.open, code_addr: int, interface: int) -> List[int]: ...
def get_smst_list(rz: rzpipe.open) -> List[int]: ...
def find_handler_register_service(insns: List[Dict[str, Any]]) -> Optional[int]: ...
def get_child_sw_smi_handler_bb(rz: rzpipe.open, insns: List[Dict[str, Any]]) -> Optional[ChildSwSmiHandler]: ...
def get_child_sw_smi_handlers(rz: rzpipe.open, smst_list: List[int]) -> List[ChildSwSmiHandler]: ...
def get_smi_handlers(rz: rzpipe.open) -> List[SmiHandler]: ...
